module Graphics.Exif (ExifData, loadFromFile, loadFromData, getTag) where

import Foreign
import Foreign.C

#include <exif-data.h>

data ExifData_ = ExifData_

newtype ExifData  = ExifData (ForeignPtr ExifData_)

data ExifContent_ = ExifContent_
type ExifContent  = Ptr ExifContent_

data ExifEntry_ = ExifEntry_
type ExifEntry  = Ptr ExifEntry_

type ExifIfd = CInt
type ExifTag = CInt

foreign import ccall "exif-data.h exif_data_new_from_file" 
	exif_data_new_from_file :: CString -> IO (Ptr ExifData_)

foreign import ccall "exif-data.h exif_data_new_from_data" 
	exif_data_new_from_data :: Ptr CUChar -> CUInt -> IO (Ptr ExifData_)

foreign import ccall "exif-data.h &exif_data_free" 
	p_exif_data_free :: FunPtr (Ptr ExifData_ -> IO ())

foreign import ccall "exif-data.h exif_content_get_entry"
        exif_content_get_entry :: Ptr ExifContent_ -> ExifTag 
                               -> IO (Ptr ExifEntry_)

foreign import ccall "exif-data.h exif_entry_get_value"
        exif_entry_get_value :: Ptr ExifEntry_ -> Ptr CChar 
                             -> CUInt -> IO (Ptr CChar)

foreign import ccall "exif-data.h exif_tag_from_name"
        exif_tag_from_name :: CString -> IO ExifTag

exif_data_get_content :: Ptr ExifData_ -> ExifIfd -> IO (Ptr ExifContent_)
exif_data_get_content d i = do
                            let arr = #{ptr ExifData, ifd} d
                            peekElemOff arr (fromIntegral i)

ifd_0 :: ExifIfd
ifd_0 = #{const EXIF_IFD_0}

ifd_1 :: ExifIfd
ifd_1 = #{const EXIF_IFD_1}

ifd_exif :: ExifIfd
ifd_exif = #{const EXIF_IFD_EXIF}

ifd_gps :: ExifIfd
ifd_gps = #{const EXIF_IFD_GPS}

ifd_interoperability :: ExifIfd
ifd_interoperability = #{const EXIF_IFD_INTEROPERABILITY}

mkExifData :: Ptr ExifData_ -> IO ExifData
mkExifData = fmap ExifData . newForeignPtr p_exif_data_free

loadFromFile :: FilePath -> IO ExifData
loadFromFile f = 
    do
    e <- withCString f exif_data_new_from_file
    mkExifData e

loadFromData :: Ptr () -> Int -> IO ExifData
loadFromData b l = 
    do
    e <- exif_data_new_from_data (castPtr b) (fromIntegral l)
    mkExifData e

getTag :: ExifData -> String -> IO String
getTag (ExifData d) s = 
    do
    t <- withCString s exif_tag_from_name
    withForeignPtr d (\p -> f p t ifds)
 where f _ _ [] = return ""
       f p t (i:is) = do
                      c <- exif_data_get_content p i
                      if c == nullPtr 
                       then f p t is 
                       else do
                            e <- exif_content_get_entry c t
                            if e == nullPtr 
                               then f p t is 
                               else entryGetValue e
       ifds = [ifd_0,ifd_1,ifd_exif,ifd_gps,ifd_interoperability]

entryGetValue :: ExifEntry -> IO String
entryGetValue e = allocaArray bufSize f
  where bufSize = 256
        f buf = do
                exif_entry_get_value e buf (fromIntegral bufSize)
                peekCString buf
